//---------------------------------------------------------------------------
// EntryPoint.cpp
//
// Version: 1.0.1
// Author: Moochos
//---------------------------------------------------------------------------
#include <windows.h>
#pragma hdrstop
#include "EntryPoint.h"
#include "WheelCommand.h"
//---------------------------------------------------------------------------
namespace {
//---------------------------------------------------------------------------
struct TSharedData
{
	HHOOK Hook;
	bool Enabled;
	Wheeler_TSettings Settings;
};
HINSTANCE Instance = NULL;
HHOOK Hook = NULL;
HANDLE SharedMemory = NULL;
HANDLE InstanceLock = NULL;
bool InitialInstance = false;
bool Initialize(HINSTANCE AInstance);
void Uninitialize();
bool InitializeInstanceLock();
void UninitializeInstanceLock();
bool LockInstance();
void UnlockInstance();
bool InstallHook();
void UninstallHook();
bool InitializeSharedMemory();
void UninitializeSharedMemory();
bool InitializeSharedData();
bool UpdateLocalCache();
LRESULT CALLBACK HookProc(int nCode, WPARAM wParam, LPARAM lParam);
void WMMouseWheel(MSG& AMessage);
bool GetSharedData(TSharedData& ASharedData);
bool SetSharedData(const TSharedData& ASharedData);
//---------------------------------------------------------------------------
bool Initialize(HINSTANCE AInstance)
{
	bool completed = false;
	try {
		Instance = AInstance;
		if(!InitializeInstanceLock()) {
			return false;
		}
		if(!LockInstance()) {
			return false;
		}
		try {
			if(!InitializeSharedMemory()) {
				return false;
			}
			if(InitialInstance) {
				if(!InstallHook() || !InitializeSharedData()) {
					return false;
				}
			}
			if(!UpdateLocalCache()) {
				return false;
			}
			completed = true;
		}
		__finally {
			UnlockInstance();
		}
	}
	__finally {
		if(!completed) {
			Uninitialize();
		}
	}
	return true;
}
//---------------------------------------------------------------------------
void Uninitialize()
{
	if(InitialInstance) {
		// Do need to uninitialize the shared data
		UninstallHook();
	}
	UninitializeSharedMemory();
	UninitializeInstanceLock();
}
//---------------------------------------------------------------------------
bool InitializeInstanceLock()
{
	::SetLastError(NO_ERROR);
	InstanceLock = ::CreateMutex(NULL, false, "Moochos.WheelHere.Wheeler.EntryPoint.InstanceLock");
	if(InstanceLock == NULL) {
		return false;
	}
	InitialInstance = ::GetLastError() != ERROR_ALREADY_EXISTS;
	return true;
}
//---------------------------------------------------------------------------
void UninitializeInstanceLock()
{
	if(InstanceLock == NULL) {
		return;
	}
	::CloseHandle(InstanceLock);
	InstanceLock = NULL;
}
//---------------------------------------------------------------------------
bool LockInstance()
{
	switch(::WaitForSingleObject(InstanceLock, INFINITE)) {
	case WAIT_ABANDONED:
	case WAIT_OBJECT_0:
		return true;
	default:
		return false;
	}
}
//---------------------------------------------------------------------------
void UnlockInstance()
{
	::ReleaseMutex(InstanceLock);
}
//---------------------------------------------------------------------------
bool InstallHook()
{
	Hook = ::SetWindowsHookEx(
		WH_GETMESSAGE,
		reinterpret_cast<HOOKPROC>(&HookProc),
		Instance,
		0);
	return Hook != NULL;
}
//---------------------------------------------------------------------------
void UninstallHook()
{
	if(Hook == NULL) {
		return;
	}
	::UnhookWindowsHookEx(Hook);
	Hook = NULL;
}
//---------------------------------------------------------------------------
bool InitializeSharedMemory()
{
	SharedMemory = ::CreateFileMapping(
		reinterpret_cast<HWND>(0xFFFFFFFF),
		NULL,
		PAGE_READWRITE,
		0,
		sizeof(TSharedData),
		"Moochos.WheelHere.Wheeler.EntryPoint.SharedMemory");
	return SharedMemory != NULL;
}
//---------------------------------------------------------------------------
void UninitializeSharedMemory()
{
	if(SharedMemory == NULL) {
		return;
	}
	::CloseHandle(SharedMemory);
	SharedMemory = NULL;
}
//---------------------------------------------------------------------------
bool InitializeSharedData()
{
	TSharedData sharedData;
	sharedData.Hook = Hook;
	sharedData.Enabled = false;
	sharedData.Settings.Focus = true;
	sharedData.Settings.Scroll = true;
	return SetSharedData(sharedData);
}
//---------------------------------------------------------------------------
bool UpdateLocalCache()
{
	TSharedData sharedData;
	if(!GetSharedData(sharedData)) {
		return false;
	}
	Hook = sharedData.Hook;
	return true;
}
//---------------------------------------------------------------------------
LRESULT CALLBACK HookProc(int nCode, WPARAM wParam, LPARAM lParam)
{
	const LRESULT returnValue = ::CallNextHookEx(Hook, nCode, wParam, lParam);
		// Let other hook procedures handle the message first because this
		// procedure may clear the message.
	if(nCode == HC_ACTION && wParam == PM_REMOVE &&
		reinterpret_cast<MSG*>(lParam)->message == WM_MOUSEWHEEL) {
		WMMouseWheel(*reinterpret_cast<MSG*>(lParam));
	}
	return returnValue;
}
//---------------------------------------------------------------------------
void WMMouseWheel(MSG& AMessage)
{
	try {
		TSharedData sharedData;
		if(!GetSharedData(sharedData) || !sharedData.Enabled) {
			return;
		}
		TWheelCommand command(AMessage);
		command.Focus = sharedData.Settings.Focus;
		command.Scroll = sharedData.Settings.Scroll;
		command.Execute();
	}
	catch(...) {
		// Do nothing
	}
}
//---------------------------------------------------------------------------
bool GetSharedData(TSharedData& ASharedData)
{
	if(!LockInstance()) {
		return false;
	}
	try {
		void* memory = ::MapViewOfFile(
			SharedMemory,
			FILE_MAP_ALL_ACCESS,
			0,
			0,
			sizeof(TSharedData));
		if(memory == NULL) {
			return false;
		}
		try {
			::CopyMemory(&ASharedData, memory, sizeof(TSharedData));
		}
		__finally {
			::UnmapViewOfFile(memory);
		}
	}
	__finally {
		UnlockInstance();
	}
	return true;
}
//---------------------------------------------------------------------------
bool SetSharedData(const TSharedData& ASharedData)
{
	if(!LockInstance()) {
		return false;
	}
	try {
		void* memory = ::MapViewOfFile(
			SharedMemory,
			FILE_MAP_ALL_ACCESS,
			0,
			0,
			sizeof(TSharedData));
		if(memory == NULL) {
			return false;
		}
		try {
			::CopyMemory(memory, &ASharedData, sizeof(TSharedData));
			if(!::FlushViewOfFile(memory, 0)) {
				return false;
			}
		}
		__finally {
			::UnmapViewOfFile(memory);
		}
	}
	__finally {
		UnlockInstance();
	}
	return true;
}
//---------------------------------------------------------------------------
} // End of namespace
//---------------------------------------------------------------------------
#pragma argsused
BOOL WINAPI DllEntryPoint(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	switch(fdwReason) {
	case DLL_PROCESS_ATTACH:
		return ::DisableThreadLibraryCalls(hinstDLL) && Initialize(hinstDLL);
			// Call DisableThreadLibraryCalls to prevent the problem that the
			// thread local storage is uninitialized for over times. The
			// problem is caused by not using the dynamic runtime library
	case DLL_PROCESS_DETACH:
		Uninitialize();
		return true;
	case DLL_THREAD_ATTACH:
		return true;
	case DLL_THREAD_DETACH:
		return true;
	default:
		return true;
	}
}
//---------------------------------------------------------------------------
bool __stdcall Wheeler_IsEnabled(bool& AEnabled)
{
	TSharedData sharedData;
	if(!GetSharedData(sharedData)) {
		return false;
	}
	AEnabled = sharedData.Enabled;
	return true;
}
//---------------------------------------------------------------------------
bool __stdcall Wheeler_SetEnabled(bool AEnabled)
{
	TSharedData sharedData;
	if(!GetSharedData(sharedData)) {
		return false;
	}
	sharedData.Enabled = AEnabled;
	if(!SetSharedData(sharedData)) {
		return false;
	}
	return true;
}
//---------------------------------------------------------------------------
bool __stdcall Wheeler_GetSettings(Wheeler_TSettings& ASettings)
{
	TSharedData sharedData;
	if(!GetSharedData(sharedData)) {
		return false;
	}
	ASettings = sharedData.Settings;
	return true;
}
//---------------------------------------------------------------------------
bool __stdcall Wheeler_SetSettings(const Wheeler_TSettings& ASettings)
{
	TSharedData sharedData;
	if(!GetSharedData(sharedData)) {
		return false;
	}
	sharedData.Settings = ASettings;
	if(!SetSharedData(sharedData)) {
		return false;
	}
	return true;
}
//---------------------------------------------------------------------------

