//---------------------------------------------------------------------------
// WheelCommand.cpp
//
// Version: 1.0.2
// Author: Moochos
//---------------------------------------------------------------------------
#include <windows.h>
#pragma hdrstop
#include <windowsx.h>
#include "WheelCommand.h"
#include "WheelCommon.h"
//---------------------------------------------------------------------------
namespace Moochos { namespace WheelHere { namespace Wheeler { namespace WheelCommand {
//---------------------------------------------------------------------------
namespace {
//---------------------------------------------------------------------------
// Key values
const WORD MK_REDIRECTED = 0x8000;
const WORD MK_HANDLED = 0x4000;
//---------------------------------------------------------------------------
} // End of namespace
//---------------------------------------------------------------------------
bool TWheelCommand::ValidateMessage()
{
	FWindowSnap = TWindowSnap(GET_X_LPARAM(FMessage.lParam), GET_Y_LPARAM(FMessage.lParam));
	FMessage.hwnd = FWindowSnap.TargetWindow;

	// Do nothing if the message has been already handled
	if((LOWORD(FMessage.wParam) & MK_HANDLED) != 0) {
		ClearMessage();
		return false;
	}

	// Do nothing if the target window is unavailable
	if(FWindowSnap.TargetWindow == NULL) {
		ClearMessage();
		return false;
	}

	// Redirect the message to its thread
	if(::GetWindowThreadProcessId(FWindowSnap.TargetWindow, NULL) != ::GetCurrentThreadId()) {
		if((LOWORD(FMessage.wParam) & MK_REDIRECTED) != 0) {
			ClearMessage();
			return false;
		}
		FMessage.wParam = MAKEWPARAM(
			LOWORD(FMessage.wParam) | MK_REDIRECTED,
			HIWORD(FMessage.wParam));
		::PostMessage(FWindowSnap.TargetWindow, FMessage.message, FMessage.wParam, FMessage.lParam);
		ClearMessage();
		return false;
	}

	// Mark the message to prevent the infinite loop
	FMessage.wParam = MAKEWPARAM(
		LOWORD(FMessage.wParam) | MK_HANDLED,
		HIWORD(FMessage.wParam));

	return true;
}
//---------------------------------------------------------------------------
void TWheelCommand::PerformFocus()
{
	if(!Focus) {
		return;
	}

	// Set the foreground window to the modal window
	if(FWindowSnap.ModalWindow != NULL && FWindowSnap.ModalWindow != FWindowSnap.TopLevelWindow) {
		::ForceSetForegroundWindow(FWindowSnap.ModalWindow);
		return;
	}

	// Set the foreground window to the top-level window
	if(FWindowSnap.TopLevelWindow != NULL) {
		::ForceSetForegroundWindow(FWindowSnap.TopLevelWindow);
	}

	// Bring the MDI child window to the top
	if(FWindowSnap.MdiChildWindow != NULL) {
		::ForceSetForegroundWindow(FWindowSnap.MdiChildWindow);
	}

	// Set the keyboard focus to the target window
	if(FWindowSnap.Enabled && FWindowSnap.Visible && FWindowSnap.WantsFocus) {
		::SetFocus(FWindowSnap.TargetWindow);
	}
}
//---------------------------------------------------------------------------
void TWheelCommand::PerformScroll()
{
	// Do nothing if the target window is disabled by the modal window
	if(FWindowSnap.ModalWindow != NULL && FWindowSnap.ModalWindow != FWindowSnap.TopLevelWindow) {
		ClearMessage();
		return;
	}

	// Do nothing if the target window cannot be accessed
	if(!FWindowSnap.Enabled || !FWindowSnap.Visible) {
		ClearMessage();
		return;
	}

	// Scroll the target window
	if(!Scroll || !(FScrollSnap = TScrollSnap(FWindowSnap.TargetWindow)).CanScroll()) {
		return;
	}
	if(AttemptScroll() || EmulateScroll()) {
		ClearMessage();
		return;
	}
	ClearMessage();
}
//---------------------------------------------------------------------------
bool TWheelCommand::AttemptScroll()
{
	::SendMessage(FWindowSnap.TargetWindow, FMessage.message, FMessage.wParam, FMessage.lParam);
	return FScrollSnap != TScrollSnap(FWindowSnap.TargetWindow);
}
//---------------------------------------------------------------------------
bool TWheelCommand::EmulateScroll()
{
	if((GET_KEYSTATE_WPARAM(FMessage.wParam) & (MK_CONTROL | MK_SHIFT)) != 0) {
		return true;
	}
	FSystemSnap = TSystemSnap(true);
	return EmulateScrollTo(false) || EmulateScrollTo(true);
}
//---------------------------------------------------------------------------
bool TWheelCommand::EmulateScrollTo(bool AControl)
{
	int scrollDelta = GET_WHEEL_DELTA_WPARAM(FMessage.wParam) * -1;
	MSG message = {0};
	if(AControl) {
		if(!FScrollSnap.ControlSnap.CanScrollTo(scrollDelta >= 0)) {
			return false;
		}
		if(FScrollSnap.HorzControl) {
			message.message = WM_HSCROLL;
			message.wParam = FSystemSnap.ScrollByPage ?
				(scrollDelta < 0 ? SB_PAGELEFT : SB_PAGERIGHT) :
				(scrollDelta < 0 ? SB_LINELEFT : SB_LINERIGHT);
			message.lParam = reinterpret_cast<LPARAM>(FWindowSnap.TargetWindow);
		}
		else {
			message.message = WM_VSCROLL;
			message.wParam = FSystemSnap.ScrollByPage ?
				(scrollDelta < 0 ? SB_PAGEUP : SB_PAGEDOWN) :
				(scrollDelta < 0 ? SB_LINEUP : SB_LINEDOWN);
			message.lParam = reinterpret_cast<LPARAM>(FWindowSnap.TargetWindow);
		}
	}
	else {
		// The horizontal scroll bar can be scrolled if the virtical bar does
		// not exist
		if(FScrollSnap.VertBarSnap.Available) {
			if(!FScrollSnap.VertBarSnap.CanScrollTo(scrollDelta >= 0)) {
				return false;
			}
			message.message = WM_VSCROLL;
			message.wParam = FSystemSnap.ScrollByPage ?
				(scrollDelta < 0 ? SB_PAGEUP : SB_PAGEDOWN) :
				(scrollDelta < 0 ? SB_LINEUP : SB_LINEDOWN);
			message.lParam = 0;
		}
		else if(FScrollSnap.HorzBarSnap.Available) {
			if(!FScrollSnap.HorzBarSnap.CanScrollTo(scrollDelta >= 0)) {
				return false;
			}
			message.message = WM_HSCROLL;
			message.wParam = FSystemSnap.ScrollByPage ?
				(scrollDelta < 0 ? SB_PAGELEFT : SB_PAGERIGHT) :
				(scrollDelta < 0 ? SB_LINELEFT : SB_LINERIGHT);
			message.lParam = 0;
		}
		else {
			return false;
		}
	}
	int scrollCount = (FSystemSnap.ScrollByPage ? 1 : FSystemSnap.ScrollLineCount) *
		((scrollDelta < 0 ? -scrollDelta : scrollDelta) / WHEEL_DELTA);
	for(int i = 0; i < scrollCount; i++) {
		::SendMessage(FWindowSnap.TargetWindow, message.message, message.wParam, message.lParam);
	}
	::SendMessage(FWindowSnap.TargetWindow, message.message, SB_ENDSCROLL, message.lParam);
	return TScrollSnap(FWindowSnap.TargetWindow) != FScrollSnap;
}
//---------------------------------------------------------------------------
void TWheelCommand::ClearMessage()
{
	FMessage.hwnd = NULL;
	FMessage.message = WM_NULL;
	FMessage.wParam = 0;
	FMessage.lParam = 0;
}
//---------------------------------------------------------------------------
TWheelCommand::TWheelCommand(MSG& AMessage) :
	FMessage(AMessage),
	FWindowSnap(),
	FScrollSnap(),
	FSystemSnap(),
	FFocus(true),
	FScroll(true)
{
	// Do nothing
}
//---------------------------------------------------------------------------
TWheelCommand::~TWheelCommand()
{
	// Do nothing
}
//---------------------------------------------------------------------------
void TWheelCommand::Execute()
{
	if(!ValidateMessage()) {
		return;
	}
	PerformFocus();
	PerformScroll();
}
//---------------------------------------------------------------------------
}}}} // End of namespace Moochos::WheelHere::Wheeler::WheelCommand
//---------------------------------------------------------------------------

